from datetime import datetime
import logging
from os import getenv
from typing import FrozenSet
from fuzzywuzzy import fuzz

from .ResultMonad import Err


def get_logger(name: str) -> logging.Logger:
    """
    Returns a logger object. Pass the name of the logger as an argument.

    You can also pass the environment variables 'LOGGING_LEVEL' or 'LOG_LEVEL' to set the logging level. If none is passed, the default is 'INFO'.
    """

    if name is None or not isinstance(name, str):
        raise ValueError("Utils::GetLogger::Argument 'name' has to be a non None string.")

    logging_level = getenv("LOGGING_LEVEL", "")

    if not logging_level:
        logging_level = getenv("LOG_LEVEL", "INFO")

    logger = logging.getLogger(name)
    logger.setLevel(logging_level)

    if not logger.handlers:
        handler = logging.StreamHandler()
        handler.setLevel(logging_level)
        datefmt = "%Y-%m-%d %H:%M:%S"
        formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s", datefmt)
        handler.setFormatter(formatter)
        logger.addHandler(handler)

    return logger


def remove_extra_whitespace(string: str) -> str:
    """
    Remove extra whitespace from a string. This version removes all newlines, tabs, carriage returns, trailing and leading whitespace, and multiple spaces in a row.

    If the input is None or not a string, a ValueError is raised.
    """
    if string is None or not isinstance(string, str):
        raise ValueError("Utils::RemoveExtraWhiteSpace::Argument string has to be a (non None) string.")

    cleaned_string = " ".join(string.split()).strip()

    return cleaned_string


def get_timestamp() -> str:
    """
    Returns a string timestamp in the format YYYYMMDD_HHMMSS.
    """
    dt = datetime.now().strftime("%Y%m%d_%H%M%S")

    return dt


def fuzzy_match_score(
    str1: str,
    str2: str,
) -> int:
    """
    Returns a fuzzy match score between two strings.

    If any of the inputs is None or not a string, a ValueError is raised.
    """

    if str1 is None or str2 is None or not isinstance(str1, str) or not isinstance(str2, str):
        raise ValueError(f"Utils::FuzzyMatchScore::Arguments have to be non None strings. Got:\n'{str1}'\n'{str2}'")

    score = fuzz.token_sort_ratio(str1, str2)

    if not isinstance(score, int):
        raise ValueError(
            f"Utils::FuzzyMatchScore::The score returned by the fuzzywuzzy library is not an integer. Got: {score}"
        )

    return score


def lginf(frame: str, msg: str, logger: logging.Logger) -> None:
    """
    Log an info message in a standard format.
    """
    logger.info(f"{frame}\n\t{msg}")


def handle_error(frame: str, msg: str, logger: logging.Logger, code: int) -> Err:
    """
    Handle errors. This function returns an Err object with the message and code.
    """
    message = f"{frame}\n\t{msg}"
    logger.error(message)
    return Err(message=message, code=code)


def handle_unexpected_exception(msg: str, logger: logging.Logger, code: int = -1) -> Err:
    """
    Handle unexpected exceptions. This function logs the message and returns an Err object with the message and code.
    """
    message = f"Unexpected exception!\n{msg}"
    logger.error(message)
    return Err(message=message, code=code)


def pretty_format_frozenset(fs: FrozenSet[object]) -> str:
    """
    Pretty format a FrozenSet object, to be used in logging or printing.
    """
    if fs is None or fs == frozenset():
        return ""
    return ", ".join(sorted([f"{item}" for item in fs]))


def dump_frozenset(fs: FrozenSet[object]) -> list[str]:
    """
    Dump a FrozenSet object to a list of strings.
    """
    if fs is None or fs == frozenset():
        return []
    return sorted([f"{item}" for item in fs])



from dataclasses import dataclass
from typing import Callable


@dataclass(slots=True, frozen=True)
class Some[T]:
    """
    Model for the success case of a function in this project.

    Attributes
    ----------
    value : T
        The data returned by the function.
    """

    value: T


@dataclass(slots=True, frozen=True)
class Nothing:
    """
    Model for the error case of a function in this project.

    Attributes
    ----------
    message : str
        A message describing the error.
    code : int
        A code that can be used to handle different error cases.
    """

    pass


type Maybe[T] = Some[T] | Nothing


def mmap[T, U](f: Callable[[T], U], maybe: Some[T] | Nothing) -> Some[U] | Nothing:
    """
    Map a function over a Maybe.

    Parameters
    ----------
    f : Callable[[T], U]
        The function to map over the Maybe.
    maybe : Maybe[T]
        The Maybe to map the function over.

    Returns
    -------
    Maybe[U]
    """
    match maybe:
        case Some(value=value):
            return Some(value=f(value))
        case Nothing():
            return maybe
        case _:
            raise ValueError("Invalid Maybe type.")


def mbind[T, U](f: Callable[[T], Some[U] | Nothing], maybe: Some[T] | Nothing) -> Some[U] | Nothing:
    """
    Bind a function over a Maybe.

    Parameters
    ----------
    f : Callable[[T], Some[U] | Nothing]
        The function to bind over the Maybe.
    maybe : Maybe[T]
        The Maybe to bind the function over.

    Returns
    -------
    Maybe[U]
    """
    match maybe:
        case Some(value):
            return f(value)
        case Nothing():
            return maybe
        case _:
            raise ValueError("Invalid Maybe type.")


def munwrap[T](maybe: Some[T] | Nothing) -> T:
    """
    Unwrap a Maybe.

    Parameters
    ----------
    maybe : Maybe[T]
        The Maybe to unwrap.

    Returns
    -------
    T
        The value of the Maybe.
    """
    match maybe:
        case Some(value):
            return value
        case Nothing():
            raise ValueError("Cannot unwrap a Nothing.")
        case _:
            raise ValueError("Invalid Maybe type.")


def munwrap_or[T](maybe: Some[T] | Nothing, default: T) -> T:
    """
    Unwrap a Maybe or return a default value.

    Parameters
    ----------
    maybe : Maybe[T]
        The Maybe to unwrap.
    default : T
        The default value to return if the Maybe is a Nothing.

    Returns
    -------
    T
        The value of the Maybe or the default value.
    """
    match maybe:
        case Some(value):
            return value
        case Nothing():
            return default
        case _:
            raise ValueError("Invalid Maybe type.")



from typing import Callable, Generator, Iterator


class ReusableGenerator[T]():
    """
    Adapter class to allow consuming a generator more than once. Usage example:

    ```python
    def my_generator() -> Generator[int, None, None]:
        yield 1
        yield 2
        yield 3

    adapter = GeneratorAdapter(my_generator)

    for item in adapter:
        print(item)

    for item in adapter:
        print(item)
    ```
    """

    def __init__(self, iterator_factory: Callable[[], Generator[T, None, None]]) -> None:
        self.iterator_factory = iterator_factory

    def __iter__(self) -> Iterator[T]:
        return self.iterator_factory()



### TO DELETE IF THIS EVER GETS IMPLEMENTED IN THE MAIN LIBRARY ###
import sys

sys.path.append('..')
### ###

from typing import Callable

from src.aletk.ResultMonad import Err, Ok


class pipe[T]:
    """
    A pipe object that can be used to chain functions together with the `>>` operator.

    ### Example

    ```python
    from aletk import pipe

    def add_one(x: int) -> int:
        return x + 1

    def multiply_by_two(x: int) -> int:
        return x * 2

    result = pipe(1) >> add_one >> multiply_by_two
    print(result.output)  # 4
    ``
    """

    def __init__(self, value: T):
        self.value = value

    @property
    def output(self) -> T:
        return self.value

    def __rshift__[
        U
    ](self, func: Callable[[T | 'pipe[T]'], U | Ok[U]]) -> 'pipe[U]' | U | 'pipe[T]' | Err | 'pipe[Err]' | Ok[U]:
        if func is pipe_out:
            # If the function is `pipe_out`, call it directly on `self`
            return func(self)

        if isinstance(self.value, Err):
            # Propagate Err without calling the function
            return self

        if isinstance(self.value, Ok):
            # Unwrap the Ok value
            unwrapped_value = self.value.out
        else:
            # Regular value
            unwrapped_value = self.value

        # Call the function
        result = func(unwrapped_value)

        if isinstance(result, Err):
            # Propagate Err if returned by the function
            return pipe(result)
        elif isinstance(result, Ok):
            # Wrap Ok result in pipe
            return pipe(result)  # type: ignore
        else:
            # Wrap regular result in pipe
            return pipe(result)

    def __or__[U](self, func: Callable[[T | 'pipe[T]'], U]) -> 'pipe[U]' | U:
        if func is pipe_out:
            return func(self)
        return pipe(func(self.value))

    def __repr__(self) -> str:
        return f'{self.value}'


def pipe_out[T](pipe_result: pipe[T] | pipe[Ok[T]] | T) -> T:
    if isinstance(pipe_result, pipe):
        if isinstance(pipe_result.output, Ok):
            return pipe_result.output.out

        return pipe_result.output  # Return the unwrapped result
    return pipe_result
